-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}

local regenerationRate = 0.0005
local regenerationEnabled = false
local repairErrorMargin = 0.002
local regenBeamStates = {}
local regenProgress = 0
local regenCancelled = false
local state = {}

local function cancelRegenerate()
  regenerationEnabled = false
  regenCancelled = true
  M.requestState()
end

local function regenerate()
  if regenerationEnabled then return end
  regenerationEnabled = true
  regenCancelled = false
  
  --no idea, copied from old mod
  local i,beam
  for i, beam in pairs (v.data.beams) do
      obj:setBeamSpringDamp(beam.cid, beam.beamSpring or -1, beam.beamDamp or -1, beam.springExpansion or -1, beam.dampExpansion or -1)
  end
  
  --reset beam states
  regenBeamStates = {}
end

local function scaleThresh()
	return regenerationRate * bullettime.get()
end

local function validRegenPart(partName)
	local isValid = true
  local po = partName:lower()
	if string.find(po, "driveshaft") or string.find(po,"coilover") or string.find(po,"suspension") or  string.find(po,"steering") or  string.find(po,"swaybar") or  string.find(po,"brake") or string.find(po,"spring") or string.find(po,"wheel") then isValid = false end
	return isValid
end

local function getBeamPartOrigin(b, getEndNode)
	local x = nil
  
  if not getEndNode then
    x = v.data.nodes[b.id1].partOrigin
  else
    x = v.data.nodes[b.id2].partOrigin
  end
  
	if x == nil then x = "" end
	return x
end

local function doRegen()
		local timeScaledRate = scaleThresh()
		local numFull = 0
		local numTotal = 0
		for k,beam in pairs(v.data.beams) do
      if regenBeamStates[beam.cid] ~= nil and regenBeamStates[beam.cid] == true then numTotal = numTotal + 1 numFull = numFull + 1 goto next end --already done!
      
			local isValid = validRegenPart(getBeamPartOrigin(beam,false)) and validRegenPart(getBeamPartOrigin(beam,true)) and (beam.beamType == "|NORMAL" or beam.beamType == "|BOUNDED")
			if not isValid or obj:beamIsBroken(beam.cid) then goto next end --skip!
           
      --get beam length, check if this is a 'weird' beam?
      local currentLengthRatio = obj:getBeamLengthRefRatio(beam.cid)
      if currentLengthRatio < 0.001 then goto next end
      
      --add one to the total calculations
		  numTotal = numTotal + 1
      
      --change length if we need to
      local currentLength = obj:getBeamLength(beam.cid)
      local targetLength = (currentLength / currentLengthRatio)
      local remainingLengthChange = math.abs(targetLength - currentLength)
      local lengthScaledRate = timeScaledRate / targetLength
      
      if  remainingLengthChange < repairErrorMargin then
        numFull = numFull + 1
        regenBeamStates[beam.cid] = true
      else
        if currentLength < targetLength then
          obj:setBeamLengthRefRatio(k,currentLengthRatio + lengthScaledRate)
        else
          obj:setBeamLengthRefRatio(k,currentLengthRatio - lengthScaledRate)
        end
      end
			
      ::next::
		end
		
    --we're done?
    regenProgress = numFull / numTotal
    if numFull == numTotal then 
      M.requestState()
      regenerationEnabled = false 
      material.reset()
    end
end

local function onReset()
  regenerationEnabled = false
  regenCancelled = false
  regenBeamStates = {}
end

local function updateGFX(dt)
  if not regenerationEnabled then
    return
  end

  doRegen()
  M.requestState()
end


local function requestState()
  state.progress = regenProgress
  state.enabled = regenerationEnabled
  
  if regenerationEnabled then
    state.status = "Regeneration in progress..."
    if regenProgress > 0.99 then state.status = state.status .. " (Sit tight)" end
  elseif regenCancelled then
    state.status = "Regeneration cancelled."
  else
    state.status = "Regeneration complete!"
  end
  
  if not playerInfo.firstPlayerSeated then return end
  guihooks.trigger('BeamRegeneratorState', state)
end

-- public interface
M.regenerate = regenerate
M.cancelRegenerate = cancelRegenerate
M.onReset   = onReset
M.updateGFX = updateGFX
M.requestState = requestState

return M
